/*
 * Program:
 *		Qnx2fsys
 *
 * Description:
 *		A sample RESOURCE manager that reads QNX2 partitions
 *
 * Author:
 *		Peter van der Veen
 *
 * Bugs:
 *		- Relitave and reverse seeking is not working correctly
 *		- You can not currently specify a block special device to
 *		  assign to a drive.
 */

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <sys/psinfo.h>
#include <sys/sys_msg.h>
#include <sys/kernel.h>
#include <sys/sendmx.h>
#include <sys/prfx.h>
#include <sys/io_msg.h>
#include <sys/fsys_msg.h>
#include <sys/dev_msg.h>
#include <sys/fd.h>
#include <sys/fsys.h>
#include <sys/disk.h>
#include <sys/name.h>
#include <dirent.h>
#include <time.h>
#include <fnmatch.h>

/* so that this can be compiled on pre-4.10 systems */
#ifndef _FSYS_OLD_CHDIR
#define _FSYS_OLD_CHDIR _FSYS_CHDIR
#endif

#ifdef GOOD__USAGE   /* to be used when program is complete */
%C [-S] [-m] [qnx2_drive_name=qnx4_drive_name]*  &
%C -i [-n node] [qnx2_drive_path]*
%C -x [-n node]
 where:
   -S              - Scan '/dev' directory for drives to adopt
   -i              - Get information about an adopted QNX 2 drive
   -x              - Cause the Qnx2fsys server to exit
   -m              - do not mount the drives. Display what would be mounted.
   -n              - Node number of Qnx2fsys manager
   qnx2_drive_name - QNX 2.xx drive number ( 1-8 )
   qnx4_drive_name - name of the QNX 4.X drive/file to open
   qnx2_drive_path - name of adopted QNX 2 drive ( /qnx2/3, /qnx2/4 )

 defaults:
   qnx2_name - /qnx2/1 (for 1=), /qnx2/2 (for 2=), /qnx2/3 (for 3=) etc.

 examples:
  Qnx2fsys  1=/dev/fd0  3=/dev/hd0t7 &
    would adopt '/dev/fd0' as '/qnx2/1' and '/dev/hd0t7' as '/qnx2/3'

  Qnx2fsys  &
  Qnx2fsys -S &
    would scan the '/dev' directory and mount corresponding drives
#endif

#ifdef __USAGE
%C [-S] [-m]  &
 where:
   -S              - Scan '/dev' directory for drives to adopt
   -m              - do not mount the drives. Display what would be mounted.

 defaults:
   qnx2_name - /qnx2/3 (for qnx), /qnx2/4 (for qny), /qnx2/5 (for qnz) etc.

 examples:

  Qnx2fsys -S &
    would scan the '/dev' directory and mount corresponding drives
#endif

#define		SECONDS_BETWEEN_1970_AND_1980	315532800

#define		PFLAGS	(_PPF_SERVER|_PPF_PRIORITY_REC)

struct	_sysmsg_version_reply	version_stamp = {
		"Qnx2fsys",
		__DATE__,
		001,
		'E',
		0
	};

union {
	msg_t	type;
	struct receive_message {
		struct _sysmsg_hdr	hdr;
		union {
			struct _sysmsg_death	death;
			struct _sysmsg_signal	signal;
			struct _sysmsg_trace	trace;
			struct _sysmsg_version	version;
		} un;
	} sys;
	struct _io_open		open;
	struct _io_write	write;
	struct _io_read		read;
	struct _io_fstat	fstat;
	struct _io_lseek	lseek;
	struct _io_rewinddir rwd;
	struct _io_close	close;
	struct _io_readdir	readdir;
	struct _io_dup		dup;
	struct _dev_tcgetattr tcga;
	struct _fsys_xstat	xstat;
	struct _fsys_disk_space	space;
	struct _fsys_get_entry	entry;
	char buff[9999];
} rmsg;

union reply_message {
	msg_t						status;
	struct _sysmsg_hdr_reply	sys;
	struct _io_open_reply		open;
	struct _io_write_reply		write;
	struct _io_read_reply		read;
	struct _io_fstat_reply		fstat;
	struct _io_lseek_reply		lseek;
	struct _io_rewinddir_reply	rwd;
	struct _io_close_reply		close;
	struct _io_readdir_reply	readdir;
	struct _io_dup_reply		dup;
	struct _dev_tcgetattr_reply	tcga;
	struct _fsys_xstat_reply	xstat;
	struct _fsys_get_mount_reply	mount;
	struct _fsys_disk_space_reply	space;
	struct _fsys_get_entry_reply	entry;
	char buff[9999];
} smsg;

#define Q2_FILE_MODIFIED	0x10
#define Q2_USED				0x40
#define Q2_FILE_BUSY		0x80

#define Q2_READ				0x01
#define Q2_WRITE			0x02
#define Q2_APPEND			0x04
#define Q2_CREATE			0x04
#define Q2_EXECUTE			0x08
#define Q2_BLOCK			0x08
#define Q2_MODIFY			0x10
#define Q2_DIRECTORY		0x20

#define Q2_NCPFN 16
struct dir_entry {
	char fstat;
	long ffirst_xtnt;
	long flast_xtnt;
	long fnum_blks;
	unsigned short fnum_xtnt;
	unsigned char fowner;
	unsigned char fgroup;
	unsigned short fnum_chars_free; /* No. unused chars in last block */
	long fseconds;
	unsigned char ftype;
	unsigned char fgperms;
	unsigned char fperms;
	unsigned char fattr;
	unsigned short fdate[2];
	char fname[Q2_NCPFN + 1];
} dir_entry;

struct xtnt_hdr {
	long prev_xtnt;
	long next_xtnt;
	long size_xtnt;
	long bound_xtnt;
};

struct drive {
	struct drive	*next;
	int				fd;
	struct dirent	dir;
} *drives;

struct ocb {
	struct ocb		*next;
	struct ocb		*prev;
	struct fptr		*fp;
	pid_t			pid;
	int				fd;
} *top_ocb;

struct fptr {
	struct drive	*drive;
	int				count;
	int				lvl;
	long int		blk;
	int				off;
	long int		lxblk;
	long int		xoff;
	struct dir_entry de;
	struct xtnt_hdr	xh;
	char			buff[_BLOCK_SIZE];
};
	
struct _mxfer_entry mx_entry[2];

dev_t devno;

struct dirent dirbuf[_DIRBUF];
char buff[4096];

char path[PATH_MAX] = "/dev";
char qnx2path[] = "/qnx2";
char name[] = "qnx/qnx2fsys";

time_t odtond(unsigned short date[2]) {
	struct tm tp;

	tp.tm_year = (date[0] >> 12) + 80;
	tp.tm_mon = ((date[0] >> 8) & 0x0f) - 1;
	tp.tm_mday = (date[0] >> 1) & 0x3f;
	tp.tm_hour = date[1] / 3600;
	tp.tm_min = (date[1] & 3600) / 60;
	tp.tm_sec = date[1] % 60;
	return mktime(&tp);
}

struct fptr *qnx2_open(pid_t pid, int fd);
struct ocb *find(pid_t pid, int fd);
void load_stat(struct drive *drvp, struct stat *stat, struct dir_entry *de);

void main (int argc, char *argv[]) {
	pid_t			pid;
	struct fptr		*fp;
	struct ocb		*ocb;
	int				num;
	int				i;
	short int		tmp;

	if(init(argc, argv)) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}
	if(qnx_pflags(PFLAGS, PFLAGS, NULL, NULL)) {
		perror("Error setting flags");
		exit(EXIT_FAILURE);
	}
	if(qnx_prefix_attach(qnx2path, NULL, 0)) {
		perror("Can't attach prefix");
		exit(EXIT_FAILURE);
	}

	for(;;) {
		pid = Receive(0, &rmsg, sizeof(rmsg));
		switch(rmsg.type) {
		case _SYSMSG:
			switch(rmsg.sys.hdr.subtype) {
			case _SYSMSG_SUBTYPE_DEATH:
				smsg.sys.status = EOK;
				Reply(pid, &smsg, sizeof(smsg.sys));
				break;
			case _SYSMSG_SUBTYPE_SIGNAL:
				smsg.sys.status = EOK;
				Reply(pid, &smsg, sizeof(smsg.sys));
				break;
			case _SYSMSG_SUBTYPE_TRACE:
				smsg.sys.status = ENOSYS;
				Reply(pid, &smsg, sizeof(smsg.sys));
				break;
			case _SYSMSG_SUBTYPE_VERSION:
				smsg.sys.status = EOK;
				_setmx(&mx_entry[0], &smsg, sizeof(smsg.sys));
				_setmx(&mx_entry[1], &version_stamp, sizeof(version_stamp));
				Replymx(pid, 2, &mx_entry);
				break;
			default:
				smsg.sys.status = ENOSYS;
				Reply(pid, &smsg, sizeof(smsg.sys));
				break;
			}
			break;
		case _FSYS_REMOVE:
			smsg.open.status = EROFS;
			Reply(pid, &smsg, sizeof(smsg.open));
			break;
		case _DEV_TCGETATTR:
			smsg.tcga.status = ENOTTY;
			Reply(pid, &smsg, sizeof(smsg.tcga));
			break;
		case _FSYS_OLD_CHDIR:
			if(qnx2_open(pid, -1))
				qnx2_close(pid, -1);
			Reply(pid, &smsg, sizeof(smsg.open));
			break;
#ifdef _NOT_DONE_YET
		case _FSYS_GET_MOUNT:
/*
			printf("io_open fd=%d\n", rmsg.open.fd);
			printf("\tpath='%s'\n", rmsg.open.path);
			printf("\tfd=%d\n", rmsg.open.fd);
			printf("\tunit=%d\n", rmsg.open.unit);
			printf("\tnid=%d\n", rmsg.open.nid);
			printf("\tpid=%d\n", rmsg.open.pid);
			printf("\tsflag=%d\n", rmsg.open.sflag);
			printf("\tmode=%d\n", rmsg.open.mode);
*/
			smsg.mount.status = EOK;
			strcpy(smsg.mount.name, "//1/dev/hd0t7");
			Reply(pid, &smsg, sizeof(smsg.mount) + strlen(smsg.mount.name));
			break;
#endif
		case _IO_OPEN:
			if(		(rmsg.open.oflag & O_ACCMODE) == O_WRONLY ||
					(rmsg.open.oflag & O_ACCMODE) == O_RDWR ||
					rmsg.open.oflag & O_CREAT ||
					rmsg.open.oflag & O_TRUNC) {
				smsg.open.status = EROFS;
				Reply(pid, &smsg, sizeof(smsg.open));
				break;
			}
			qnx2_open(pid, rmsg.open.fd);
			Reply(pid, &smsg, sizeof(smsg.open));
			break;
		case _IO_REWINDDIR:
			if(fp = find(pid, rmsg.rwd.fd)->fp) {
				fp->blk = 0L;
				smsg.rwd.status = EOK;
			} else {
				smsg.rwd.status = EBADF;
			}
			Reply(pid, &smsg, sizeof(smsg.rwd));
			break;
		case _IO_HANDLE:
			switch(rmsg.open.oflag) {
			case _IO_HNDL_INFO:
				qnx2_open(pid, rmsg.open.fd);
				Reply(pid, &smsg, sizeof(smsg.open));
				break;
			case _IO_HNDL_RDDIR:
				qnx2_open(pid, rmsg.open.fd);
				Reply(pid, &smsg, sizeof(smsg.open));
				break;
			case _IO_HNDL_LOAD:
			case _IO_HNDL_CLOAD:
				smsg.open.status = EACCES;
				Reply(pid, &smsg, sizeof(smsg.open));
				break;
			default:
				smsg.open.status = ENOSYS;
				Reply(pid, &smsg, sizeof(smsg.open));
				break;
			}
			break;
#ifdef _NOT_DONE_YET
		case _FSYS_DISK_SPACE:
			if(fp = find(pid, rmsg.space.fd)->fp) {
				smsg.space.status = EOK;
				smsg.space.zero = 0;
				smsg.space.free_blocks = 100000L;
				smsg.space.total_blocks = 200000L;
				/* grab code from query.c and check bitmap */
			} else {
				smsg.entry.status = EBADF;
			}
			Reply(pid, &smsg, sizeof(smsg.space));
			break;
		case _FSYS_DISK_GET_ENTRY:
			if(fp = find(pid, rmsg.entry.fd)->fp) {
				smsg.entry.status = EOK;
				smsg.entry.zero = 0;
				strcpy(smsg.entry.data, "//1/dev/hd0t7");
				Reply(pid, &smsg, sizeof(smsg.entry) + strlen(smsg.entry.data));
			} else {
				smsg.entry.status = EBADF;
				Reply(pid, &smsg, sizeof(smsg.entry));
			}
			break;
#endif
		case _FSYS_XSTAT:
			if(fp = find(pid, rmsg.xstat.fd)->fp) {
				memset(&smsg.xstat.xstat, 0, sizeof(struct _fsys_xstat));
				if(fp->lvl <= 1) {
					memcpy(&smsg.xstat.xstat, &fp->drive->dir.d_stat, sizeof(struct stat));
					smsg.xstat.xstat.st_first_xtnt.xtnt_blk =  1L;
					smsg.xstat.xstat.st_num_xtnts = 1;
					smsg.xstat.xstat.st_first_xtnt.xtnt_size =
						(smsg.xstat.xstat.st_size + _BLOCK_SIZE - 1) / _BLOCK_SIZE;
				} else {
					load_stat(fp->drive, (struct stat *)&smsg.xstat.xstat, &fp->de);
					qnx2_fget(&buff, 0, fp);
					smsg.xstat.xstat.st_first_xtnt.xtnt_blk = fp->de.ffirst_xtnt;
					smsg.xstat.xstat.st_first_xtnt.xtnt_size = fp->xh.bound_xtnt + 1;
					smsg.xstat.xstat.st_xblk = fp->xh.next_xtnt;
					smsg.xstat.xstat.st_num_xtnts = fp->de.fnum_xtnt;
				}
				smsg.xstat.status = EOK;
			} else {
				smsg.xstat.status = EBADF;
			}
			Reply(pid, &smsg, sizeof(smsg.xstat));
			break;
		case _IO_FSTAT:
			if(fp = find(pid, rmsg.fstat.fd)->fp) {
				if(fp->lvl <= 1) {
					smsg.fstat.stat = fp->drive->dir.d_stat;
				} else {
					load_stat(fp->drive, &smsg.fstat.stat, &fp->de);
				}
				smsg.fstat.status = EOK;
			} else {
				smsg.fstat.status = EBADF;
			}
			Reply(pid, &smsg, sizeof(smsg.fstat));
			break;
		case _IO_DUP:
			if(fp = find(rmsg.dup.src_pid, rmsg.dup.src_fd)->fp) {
				if(ocb = (struct ocb *)malloc(sizeof(struct ocb))) {
					ocb->pid = pid;
					ocb->fd = rmsg.dup.dst_fd;
					ocb->fp = fp;
					ocb->prev = NULL;
					if(ocb->next = top_ocb) top_ocb->prev = ocb;
					top_ocb = ocb;
					ocb->fp->count++;
					smsg.open.status = EOK;
				} else {
					smsg.open.status = ENOMEM;
				}
			} else {
				smsg.readdir.status = EBADF;
			}
			Reply(pid, &smsg, sizeof(smsg.dup));
			break;
		case _IO_CLOSE:
			qnx2_close(pid, rmsg.close.fd);
			Reply(pid, &smsg, sizeof(smsg.close));
			break;
		case _IO_WRITE:
			smsg.write.status = EROFS;
			Reply(pid, &smsg, sizeof(smsg.write));
			break;
		case _IO_LSEEK:
			if(fp = find(pid, rmsg.lseek.fd)->fp) {
				smsg.lseek.status = EOK;
				switch(rmsg.lseek.whence) {
				case SEEK_END:
					load_stat(fp->drive, &dirbuf[0].d_stat, &fp->de);
					rmsg.lseek.offset = dirbuf[0].d_stat.st_size - rmsg.lseek.offset;
					/* fall through */
				case SEEK_SET:
					fp->blk = 0L;
					smsg.lseek.offset = rmsg.lseek.offset;
					if(rmsg.lseek.offset > 0) {
						while(rmsg.lseek.offset) {
							i = (rmsg.lseek.offset > INT_MAX) ?
								INT_MAX : (int) rmsg.lseek.offset;
							num = qnx2_fget(NULL, i, fp);
							rmsg.lseek.offset =- num;
							if(num != i) break;
						}								
					}
					break;
				case SEEK_CUR:
					if(rmsg.lseek.offset > 0) {
						while(rmsg.lseek.offset) {
							i = (rmsg.lseek.offset > INT_MAX) ?
								INT_MAX : (int) rmsg.lseek.offset;
							num = qnx2_fget(NULL, i, fp);
							rmsg.lseek.offset =- num;
							if(num != i) break;
						}								
					} else {
						smsg.lseek.status = ENOSYS;
					}
					break;
				}
			} else {
				smsg.lseek.status = EBADF;
			}
			Reply(pid, &smsg, sizeof(smsg.lseek));
			break;
		case _IO_READDIR:
			if(fp = find(pid, rmsg.fstat.fd)->fp) {
				num = rmsg.readdir.ndirs;
				if(num > _DIRBUF) num = _DIRBUF;
				if(fp->lvl) {
					if(fp->blk == 0) {
						qnx2_fget(&tmp, sizeof(short), fp);
						qnx2_fget(&tmp, sizeof(short), fp);
						if(tmp & 0x8000)
							qnx2_fget(&tmp, sizeof(short), fp);
					}
					i = 0;
					while(	i < num && 
							qnx2_fget(&dir_entry, sizeof(dir_entry), fp) == 
							sizeof(dir_entry)) {
						if(dir_entry.fstat & Q2_USED) {
							strcpy(dirbuf[i].d_name, dir_entry.fname);
							load_stat(fp->drive, &dirbuf[i].d_stat, &dir_entry);
							i++;
						}
					}
				} else {
					if(fp->blk == 0) {
						fp->drive = drives->next;
					}
					fp->blk++;
					for(i = 0; i < num && fp->drive; i++, fp->drive = fp->drive->next)
						dirbuf[i] = fp->drive->dir;
				}
				smsg.readdir.ndirs = i;
				smsg.readdir.status = EOK;
				_setmx(&mx_entry[0], &smsg, sizeof(smsg.readdir) - 1);
				_setmx(&mx_entry[1], dirbuf, sizeof(struct dirent) * i);
				Replymx(pid, 2, &mx_entry);
				break;
			}
			smsg.readdir.status = EBADF;
			Reply(pid, &smsg, sizeof(smsg.readdir));
			break;
		case _IO_READ:
			if(fp = find(pid, rmsg.read.fd)->fp) {
				num = rmsg.read.nbytes;
				if(num > sizeof(buff))
					num = sizeof(buff);
				num = qnx2_fget(&buff, num, fp);
				smsg.read.nbytes = num;
				smsg.read.status = EOK;
				_setmx(&mx_entry[0], &smsg, sizeof(smsg.read) - 1);
				_setmx(&mx_entry[1], buff, num);
				Replymx(pid, 2, &mx_entry);
			} else {
				smsg.read.status = EBADF;
				Reply(pid, &smsg, sizeof(smsg.read));
			}
			break;
		default:
			smsg.status = ENOSYS;
			Reply(pid, &smsg, sizeof(smsg));
			break;
		}
	}
}

static void load_stat(struct drive *drvp, struct stat *stat, struct dir_entry *de) {
	time_t clock;

	memset(stat, 0, sizeof(struct stat));
	stat->st_ino = de->ffirst_xtnt;
	stat->st_dev = drvp->dir.d_stat.st_dev;

	stat->st_size =	(de->fnum_blks * _BLOCK_SIZE) +_BLOCK_SIZE - 
					(de->fnum_xtnt * sizeof(struct xtnt_hdr));
	if(de->fnum_chars_free & 0x400)
		stat->st_size -= (de->fnum_chars_free & 0x1ff) ;

	if(clock = de->fseconds) {
		clock += SECONDS_BETWEEN_1970_AND_1980;
	} else {
		clock = odtond(de->fdate);
	}
	stat->st_ftime = clock;
	stat->st_mtime = clock;
	stat->st_atime = clock;
	stat->st_ctime = clock;
	if(de->fattr & Q2_DIRECTORY) {
		stat->st_nlink = 2;
		stat->st_mode |= S_IFDIR;

		if(!(de->fattr & Q2_BLOCK))
			stat->st_mode |= S_IRUSR|S_IXUSR;
		if(!(de->fgperms & Q2_BLOCK))
			stat->st_mode |= S_IRGRP|S_IXGRP;
		if(!(de->fperms & Q2_BLOCK))
			stat->st_mode |= S_IROTH|S_IXOTH;
	} else {
		stat->st_nlink = 1;
		stat->st_mode |= S_IFREG;

		if(de->fattr & Q2_READ)
			stat->st_mode |= S_IRUSR;
		if(de->fgperms & Q2_READ)
			stat->st_mode |= S_IRGRP;
		if(de->fperms & Q2_READ)
			stat->st_mode |= S_IROTH;

		if(de->fattr & Q2_WRITE)
			stat->st_mode |= S_IWUSR;
		if(de->fgperms & Q2_WRITE)
			stat->st_mode |= S_IWGRP;
		if(de->fperms & Q2_WRITE)
			stat->st_mode |= S_IWOTH;

		if(de->fattr & Q2_EXECUTE)
			stat->st_mode |= S_IXUSR;
		if(de->fgperms & Q2_EXECUTE)
			stat->st_mode |= S_IXGRP;
		if(de->fperms & Q2_EXECUTE)
			stat->st_mode |= S_IXOTH;
	}
		
	if(de->fstat & Q2_USED)
		stat->st_status |= _FILE_USED;
	if(de->fstat & Q2_FILE_MODIFIED)
		stat->st_status |= _FILE_MODIFIED;
	if(de->fstat & Q2_FILE_BUSY)
		stat->st_status |= _FILE_BUSY;
}

static int qnx2_fget(void *data, int size, struct fptr *fp) {
	int num = 0;
	int i;
	long l;
	char *p = (char *)data;

	if(fp->blk == -1)
		return 0;

	if(!fp->blk) {
		fp->blk = fp->de.ffirst_xtnt;
		if(block_read(fp->drive->fd, fp->blk, 1, fp->buff) != 1) {
			fp->blk = -1;
			return num;
		}
		memcpy(&fp->xh, fp->buff, sizeof(struct xtnt_hdr));
		fp->off = sizeof(struct xtnt_hdr);
		fp->lxblk = fp->blk + fp->xh.bound_xtnt;
		fp->xoff = 0L;
	}

	while(size) {
		l = fp->xh.size_xtnt - fp->xoff;
		i = l < INT_MAX ? (int)l : INT_MAX;
		if(i > _BLOCK_SIZE - fp->off)
			i = _BLOCK_SIZE - fp->off;
		if(i > size)
			i = size;
		if(i > 0) {
			if(data)
				memcpy(&p[num], &fp->buff[fp->off], i);
			size -= i;
			num += i;
			fp->off += i;
			fp->xoff += i;
		}
		if(i <= 0) {
			if(fp->blk == fp->lxblk) {
				if(fp->blk = fp->xh.next_xtnt) {
					if(block_read(fp->drive->fd, fp->blk, 1, fp->buff) != 1) {
						fp->blk = -1;
						return num;
					}
					memcpy(&fp->xh, fp->buff, sizeof(struct xtnt_hdr));
					fp->off = sizeof(struct xtnt_hdr);
					fp->lxblk = fp->blk + fp->xh.bound_xtnt;
					fp->xoff = 0L;
				} else {
					fp->blk = -1;
					return(num);
				}
			} else {
				fp->blk++;
				if(block_read(fp->drive->fd, fp->blk, 1, fp->buff) != 1) {
					fp->blk = -1;
					return num;
				}
				fp->off = 0;
			}
		}
	}	
	return num;
}

static struct ocb *find(pid_t pid, int fd) {
	struct ocb *ocb;

	for(ocb = top_ocb; ocb; ocb = ocb->next) {
		if(ocb->pid == pid && ocb->fd == fd) {
			return ocb;
		}
	}
	return NULL;
}

static void qnx2_close(pid_t pid, int fd) {
	struct ocb *ocb;

	if(ocb = find(pid, fd)) {
		if(ocb->prev == NULL && ocb->next == NULL) {
			top_ocb = NULL;
		} else {
			if(ocb == top_ocb) top_ocb = ocb->next;
			if(ocb->prev) ocb->prev->next = ocb->next;
			if(ocb->next) ocb->next->prev = ocb->prev;
		}
		if(--ocb->fp->count == 0) {
			free(ocb->fp);
		}
		free(ocb);
		smsg.close.status = EOK;
	} else {
		smsg.close.status = EBADF;
	}
}

static struct fptr *qnx2_open(pid_t pid, int fd) {
	struct fptr *fp;
	struct ocb *ocb;
	struct drive *cd;
	short	tmp;
	char 	*p;

	qnx2_close(pid, fd);

	strtok(p = rmsg.open.path, "/");
	for(cd = drives; cd; cd = cd->next) {
		if(!strcmp(p, cd->dir.d_name))
			break;
	}
	if(!cd) {
		smsg.open.status = ENOENT;
		return NULL;
	}

	if(!(fp = (struct fptr *)malloc(sizeof(struct fptr)))) {
		smsg.open.status = ENOMEM;
		return NULL;
	}
	memset(fp, 0, sizeof(struct fptr));

	if(!(ocb = (struct ocb *)malloc(sizeof(struct ocb)))) {
		smsg.open.status = ENOMEM;
		return NULL;
	}
	memset(ocb, 0, sizeof(struct ocb));

	smsg.open.status = EOK;

	ocb->fp = fp;
	ocb->pid = pid;
	ocb->fd = fd;
	fp->drive = cd;	
	fp->count = 1;
	fp->de.ffirst_xtnt = 1L;
	if(cd->dir.d_stat.st_ino) {
		qnx2_fget(buff, 0x44 - sizeof(struct xtnt_hdr), fp);
		qnx2_fget(&fp->de, sizeof(struct dir_entry), fp);
		fp->blk = 0L;
		fp->lvl++;
		while(p = strtok(NULL, "/")) {
			if(!(fp->de.fattr & Q2_DIRECTORY)) {
				smsg.open.status = ENOTDIR;
				break;
			}
			fp->lvl++;
			qnx2_fget(&tmp, sizeof(short), fp);
			qnx2_fget(&tmp, sizeof(short), fp);
			if(tmp & 0x8000)
				qnx2_fget(&tmp, sizeof(short), fp);
			while(qnx2_fget(&dir_entry, sizeof(dir_entry), fp) == sizeof(dir_entry)) {
				if(dir_entry.fstat & Q2_USED) {
					if(!strcmp(p, dir_entry.fname)) {
						fp->de = dir_entry;
						fp->blk = 0L;
						break;
					}
				}
			}
			if(fp->blk) {
				smsg.open.status = EACCES;
					break;
			}
		}
	}
	if(smsg.open.status != EOK) {
		free(fp);
		free(ocb);
		return NULL;
	}
	if(ocb->next = top_ocb) top_ocb->prev = ocb;
	top_ocb = ocb;
	return ocb->fp;
} 	

static struct drive *add_drive(void) {
	struct drive *dp, *drvp;

	if(!(dp = (struct drive *)malloc(sizeof(struct drive))))
		return NULL;
	for(drvp = drives; drvp && drvp->next; drvp = drvp->next);
	if(drvp)
		drvp->next = dp;
	else
		drives = dp;
	memset(dp, 0, sizeof(struct drive));
	return dp;
}

static int init(int argc, char *argv[]) {
	int					i;
	dev_t				devno;
	struct _psinfo		psdata;
	struct drive		*drvp;
	nid_t				node;
	time_t				clock;
	char				path[PATH_MAX];
	struct dirent		*dp;
    DIR					*dirp;
	int					fd;
	int					num;
	unsigned short		date[2];
	int					c;
	int					search = 0;
	int					display = 0;
	int					kill = 0;
	int					info = 0;
	pid_t				pid;

	node = getnid();
	i = 0;
	while((c = getopt(argc, argv, "Smixn:")) != -1) {
		switch(c) {
		case 'S':	search = 1; 			break;
		case 'm':	display = 1;			break;
		case 'x':	kill = 1;				break;
		case 'i':	info = 1;				break;
		case 'n':	node = atol(optarg);	break;
		default:	i++;					break;
		}
	}
	if(i) {
		print_usage(argv);
		exit(EXIT_FAILURE);
	}
	if(info) {
		if((pid = qnx_name_locate(node, name, 100, NULL)) == -1) {
			return -1;
		}
		printf("Info on node %ld\n", node);
		exit(EXIT_SUCCESS);
	}
	if(kill) {
		if((pid = qnx_name_locate(node, name, 100, NULL)) == -1) {
			return -1;
		}
		printf("kill server on node %ld\n", node);
		exit(EXIT_SUCCESS);
	}
	node = getnid();

	if(qnx_psinfo(PROC_PID, 0, &psdata, 0, NULL) == -1)
		return -1;
	if((i = qnx_device_attach()) == -1)
		return -1;
	devno = (dev_t) ((node << 16L) + (i << 10L));

	drvp = add_drive();
	drvp->fd = -1;
	clock = psdata.un.proc.start_time;
	drvp->dir.d_stat.st_ftime = clock;
	drvp->dir.d_stat.st_mtime = clock;
	drvp->dir.d_stat.st_atime = clock;
	drvp->dir.d_stat.st_ctime = clock;
	drvp->dir.d_stat.st_mode = S_IFDIR|S_IRUSR|S_IRGRP|S_IROTH|S_IXUSR|S_IXGRP|S_IXOTH;
	drvp->dir.d_stat.st_status = _FILE_USED;
	drvp->dir.d_stat.st_nlink = 2;
	drvp->dir.d_stat.st_dev = devno;
	drvp->dir.d_stat.st_rdev = devno++;
	drvp->dir.d_stat.st_ino = num = 0;

	optind--;
	while(++optind < argc) {
		printf("add %s\n", argv[optind]);
	}	
	if(search) {
		strcpy(path, "/dev");
		i = strlen(path);
		if(!(dirp = opendir(path))) {
			perror("opendir");
		} else {
			path[i++] = '/';
			while(dp = readdir(dirp)) {
				strcpy(&path[i], dp->d_name);
	   	
				if(!(dp->d_stat.st_status & _FILE_USED))
					stat(path, &dp->d_stat);
		
				if(S_ISBLK(dp->d_stat.st_mode)) {
					if(!fnmatch("hd*t[789]", dp->d_name, FNM_PATHNAME)) {
						if((fd = open(path, O_RDONLY)) < 0) {
							perror("open");
						} else {
							if(block_read(fd, 1L, 1, buff) != 1) {
								perror("block_read");
							} else {
								drvp = add_drive();
								drvp->fd = fd;
								sprintf(drvp->dir.d_name, "%x", num + 3);
								load_stat(drvp, &drvp->dir.d_stat, (struct dir_entry *)&buff[0x44]);
								memcpy(date, &buff[0x14], sizeof(date));
								clock = odtond(date);
								drvp->dir.d_stat.st_ftime = clock;
								drvp->dir.d_stat.st_mtime = clock;
								drvp->dir.d_stat.st_atime = clock;
								drvp->dir.d_stat.st_ctime = clock;
								drvp->dir.d_stat.st_ino = ++num;
								drvp->dir.d_stat.st_dev = devno;
								drvp->dir.d_stat.st_rdev = devno++;
								drvp->dir.d_stat.st_nlink = 2;
								drvp->dir.d_stat.st_mode = S_IFDIR|S_IRUSR|S_IRGRP|S_IROTH|S_IXUSR|S_IXGRP|S_IXOTH;
								drvp->dir.d_stat.st_status = _FILE_USED;
								printf("QNX4 device '%s' adopted as QNX2 drive '%s/%s'\n",
									path, qnx2path, drvp->dir.d_name);
							}
						}
					}
				}
			}
			closedir(dirp);
		}
	}
	if(display) {
		exit(EXIT_SUCCESS);
	}
	if(num == 0) {
		fprintf(stderr, "No drives defined\n");
		exit(EXIT_FAILURE);
	}
	drives->dir.d_stat.st_size = (long)num * sizeof(struct _dir_entry);
	if(qnx_name_attach(0, name) == -1) {
		return -1;
	}
	return 0;
}
